@@ -5,6 +5,7 @@ struct device {
 	void *private_data;
 	struct device_driver *driver;
 	struct pm_dev *pm_dev;
+	char	bus_id[20];
 };
 
 struct bus_type {
@@ -23,6 +24,9 @@ struct device_driver {
 	struct list_head device_list;
 };
 
+struct device_attribute {
+};
+
 int snd_compat_driver_register(struct device_driver *driver);
 void snd_compat_driver_unregister(struct device_driver *driver);
 
