@@ -1,5 +1,5 @@
 
-#include "hpi.h"
+#include "hpi_internal.h"
 #include "hpimsginit.h"
 
 struct hpi_handle {
@@ -493,7 +493,7 @@ u16 HPI_AdapterGetAssert(
 
 			*wAssertPresent = 1;
 
-			for (i = 0; i < STR_SIZE(HPI_STRING_LEN); i++) {
+			for (i = 0; i < HPI_STRING_LEN; i++) {
 				char c;
 				c = *Src++;
 				*Dst++ = c;
@@ -542,7 +542,7 @@ u16 HPI_AdapterGetAssertEx(
 			char *Src = (char *)hr.u.a.szAdapterAssert;
 			char *Dst = pszAssert;
 
-			for (i = 0; i < STR_SIZE(HPI_STRING_LEN); i++) {
+			for (i = 0; i < HPI_STRING_LEN; i++) {
 				char c;
 				c = *Src++;
 				*Dst++ = c;
@@ -606,6 +606,35 @@ u16 HPI_AdapterSelfTest(
 	return (hr.wError);
 }
 
+u16 HPI_AdapterDebugRead(
+	struct hpi_hsubsys *phSubSys,
+	u16 wAdapterIndex,
+	u32 dwDspAddress,
+	char *pBuffer,
+	int *dwCountBytes
+)
+{
+	struct hpi_message hm;
+	struct hpi_response hr;
+	HPI_InitMessage(&hm, HPI_OBJ_ADAPTER, HPI_ADAPTER_DEBUG_READ);
+	hm.wAdapterIndex = wAdapterIndex;
+	hm.u.ax.debug_read.dwDspAddress = dwDspAddress;
+
+	if (*dwCountBytes > sizeof(hr.u.bytes))
+		*dwCountBytes = sizeof(hr.u.bytes);
+
+	hm.u.ax.debug_read.dwCountBytes = *dwCountBytes;
+
+	HPI_Message(&hm, &hr);
+
+	if (!hr.wError) {
+		*dwCountBytes = hr.wSize - 12;
+		memcpy(pBuffer, &hr.u.bytes, *dwCountBytes);
+	} else
+		*dwCountBytes = 0;
+	return (hr.wError);
+}
+
 u16 HPI_AdapterSetProperty(
 	struct hpi_hsubsys *phSubSys,
 	u16 wAdapterIndex,
@@ -1614,6 +1643,7 @@ u16 HPI_MixerStore(
 	return (hr.wError);
 }
 
+static
 u16 HPI_ControlParamSet(
 	const struct hpi_hsubsys *phSubSys,
 	const u32 hControl,
@@ -1633,6 +1663,7 @@ u16 HPI_ControlParamSet(
 	return (hr.wError);
 }
 
+static
 u16 HPI_ControlParamGet(
 	const struct hpi_hsubsys *phSubSys,
 	const u32 hControl,
@@ -1668,6 +1699,7 @@ u16 HPI_ControlParamGet(
 #define HPI_ControlExParam2Get(s, h, a, p1, p2) \
 		HPI_ControlExParamGet(s, h, a, 0, 0, p1, p2)
 
+static
 u16 HPI_ControlQuery(
 	const struct hpi_hsubsys *phSubSys,
 	const u32 hControl,
@@ -1692,36 +1724,82 @@ u16 HPI_ControlQuery(
 	return (hr.wError);
 }
 
-#if 0
-
-u16 HPI_Tuner_QueryFrequency(
+static u16 HPI_Control_GetString(
 	const struct hpi_hsubsys *phSubSys,
-	const u32 hTuner,
-	const u32 dwIndex,
-	const u16 band,
-	u32 *pdwFreq
+	const u32 hControl,
+	const u16 wAttribute,
+	char *pszString,
+	const u32 wStringLength
 )
 {
-	return HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_FREQ,
-		dwIndex, band, pdwFreq);
+	unsigned int subStringIndex = 0, j = 0;
+	char c = 0;
+	unsigned int n = 0;
+	u16 wHE = 0;
+
+	if ((wStringLength < 1) || (wStringLength > 256))
+		return HPI_ERROR_INVALID_CONTROL_VALUE;
+	for (subStringIndex = 0; subStringIndex < wStringLength;
+		subStringIndex += 8) {
+		struct hpi_message hm;
+		struct hpi_response hr;
+
+		HPI_InitMessage(&hm, HPI_OBJ_CONTROL, HPI_CONTROL_GET_STATE);
+		u32TOINDEXES(hControl, &hm.wAdapterIndex,
+			&hm.u.c.wControlIndex);
+		hm.u.c.wAttribute = wAttribute;
+		hm.u.c.dwParam1 = subStringIndex;
+		hm.u.c.dwParam2 = 0;
+		HPI_Message(&hm, &hr);
+
+		if (subStringIndex == 0
+			&& (hr.u.cu.chars8.dwRemainingChars + 8) >
+			wStringLength)
+			return HPI_ERROR_INVALID_CONTROL_VALUE;
+
+		if (hr.wError) {
+			wHE = hr.wError;
+			break;
+		}
+		for (j = 0; j < 8; j++) {
+			c = hr.u.cu.chars8.szData[j];
+			pszString[subStringIndex + j] = c;
+			n++;
+			if (n >= wStringLength) {
+				pszString[wStringLength - 1] = 0;
+				wHE = HPI_ERROR_INVALID_CONTROL_VALUE;
+				break;
+			}
+			if (c == 0)
+				break;
+		}
+
+		if ((hr.u.cu.chars8.dwRemainingChars == 0) &&
+			((subStringIndex + j) < wStringLength) && (c != 0)) {
+			c = 0;
+			pszString[subStringIndex + j] = c;
+		}
+		if (c == 0)
+			break;
+	}
+	return wHE;
 }
 
-u16 HPI_Tuner_QueryBand(
+u16 HPI_AESEBU_Receiver_QueryFormat(
 	const struct hpi_hsubsys *phSubSys,
-	const u32 hTuner,
+	const u32 hAesRx,
 	const u32 dwIndex,
-	u16 *pwBand
+	u16 *pwFormat
 )
 {
 	u32 qr;
 	u16 err;
 
-	err = HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_BAND,
+	err = HPI_ControlQuery(phSubSys, hAesRx, HPI_AESEBURX_FORMAT,
 		dwIndex, 0, &qr);
-	*pwBand = qr;
+	*pwFormat = (u16)qr;
 	return err;
 }
-#endif
 
 u16 HPI_AESEBU_Receiver_SetFormat(
 	struct hpi_hsubsys *phSubSys,
@@ -1860,6 +1938,22 @@ u16 HPI_AESEBU_Transmitter_GetChannelStatus(
 	return HPI_ERROR_INVALID_OPERATION;
 }
 
+u16 HPI_AESEBU_Transmitter_QueryFormat(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hAesTx,
+	const u32 dwIndex,
+	u16 *pwFormat
+)
+{
+	u32 qr;
+	u16 err;
+
+	err = HPI_ControlQuery(phSubSys, hAesTx, HPI_AESEBUTX_FORMAT,
+		dwIndex, 0, &qr);
+	*pwFormat = (u16)qr;
+	return err;
+}
+
 u16 HPI_AESEBU_Transmitter_SetFormat(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -1927,6 +2021,22 @@ u16 HPI_Bitstream_GetActivity(
 	return (hr.wError);
 }
 
+u16 HPI_ChannelMode_QueryMode(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hMode,
+	const u32 dwIndex,
+	u16 *pwMode
+)
+{
+	u32 qr;
+	u16 err;
+
+	err = HPI_ControlQuery(phSubSys, hMode, HPI_CHANNEL_MODE_MODE,
+		dwIndex, 0, &qr);
+	*pwMode = (u16)qr;
+	return err;
+}
+
 u16 HPI_ChannelModeSet(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2236,6 +2346,35 @@ u16 HPI_Compander_Get(
 	return (hr.wError);
 }
 
+u16 HPI_LevelQueryRange(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	short *nMinGain_01dB,
+	short *nMaxGain_01dB,
+	short *nStepGain_01dB
+)
+{
+	struct hpi_message hm;
+	struct hpi_response hr;
+	HPI_InitMessage(&hm, HPI_OBJ_CONTROL, HPI_CONTROL_GET_STATE);
+	u32TOINDEXES(hControl, &hm.wAdapterIndex, &hm.u.c.wControlIndex);
+	hm.u.c.wAttribute = HPI_LEVEL_RANGE;
+
+	HPI_Message(&hm, &hr);
+	if (hr.wError) {
+		hr.u.c.anLogValue[0] = 0;
+		hr.u.c.anLogValue[1] = 0;
+		hr.u.c.dwParam1 = 0;
+	}
+	if (nMinGain_01dB)
+		*nMinGain_01dB = hr.u.c.anLogValue[0];
+	if (nMaxGain_01dB)
+		*nMaxGain_01dB = hr.u.c.anLogValue[1];
+	if (nStepGain_01dB)
+		*nStepGain_01dB = (short)hr.u.c.dwParam1;
+	return (hr.wError);
+}
+
 u16 HPI_LevelSetGain(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2244,6 +2383,7 @@ u16 HPI_LevelSetGain(
 {
 	struct hpi_message hm;
 	struct hpi_response hr;
+
 	HPI_InitMessage(&hm, HPI_OBJ_CONTROL, HPI_CONTROL_SET_STATE);
 	u32TOINDEXES(hControl, &hm.wAdapterIndex, &hm.u.c.wControlIndex);
 	memcpy(hm.u.c.anLogValue, anGain0_01dB,
@@ -2274,6 +2414,16 @@ u16 HPI_LevelGetGain(
 	return (hr.wError);
 }
 
+u16 HPI_Meter_QueryChannels(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hMeter,
+	u32 *pChannels
+)
+{
+	return HPI_ControlQuery(phSubSys, hMeter, HPI_METER_NUM_CHANNELS,
+		0, 0, pChannels);
+}
+
 u16 HPI_MeterGetPeak(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2583,6 +2733,22 @@ u16 HPI_ParametricEQ_GetCoeffs(
 	return (hr.wError);
 }
 
+u16 HPI_SampleClock_QuerySource(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hClock,
+	const u32 dwIndex,
+	u16 *pwSource
+)
+{
+	u32 qr;
+	u16 err;
+
+	err = HPI_ControlQuery(phSubSys, hClock, HPI_SAMPLECLOCK_SOURCE,
+		dwIndex, 0, &qr);
+	*pwSource = (u16)qr;
+	return err;
+}
+
 u16 HPI_SampleClock_SetSource(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2593,16 +2759,6 @@ u16 HPI_SampleClock_SetSource(
 		HPI_SAMPLECLOCK_SOURCE, wSource, 0);
 }
 
-u16 HPI_SampleClock_SetSourceIndex(
-	struct hpi_hsubsys *phSubSys,
-	u32 hControl,
-	u16 wSourceIndex
-)
-{
-	return HPI_ControlParamSet(phSubSys, hControl,
-		HPI_SAMPLECLOCK_SOURCE_INDEX, wSourceIndex, 0);
-}
-
 u16 HPI_SampleClock_GetSource(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2619,6 +2775,33 @@ u16 HPI_SampleClock_GetSource(
 	return (wError);
 }
 
+u16 HPI_SampleClock_QuerySourceIndex(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hClock,
+	const u32 dwIndex,
+	const u32 dwSource,
+	u16 *pwSourceIndex
+)
+{
+	u32 qr;
+	u16 err;
+
+	err = HPI_ControlQuery(phSubSys, hClock, HPI_SAMPLECLOCK_SOURCE_INDEX,
+		dwIndex, dwSource, &qr);
+	*pwSourceIndex = (u16)qr;
+	return err;
+}
+
+u16 HPI_SampleClock_SetSourceIndex(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u16 wSourceIndex
+)
+{
+	return HPI_ControlParamSet(phSubSys, hControl,
+		HPI_SAMPLECLOCK_SOURCE_INDEX, wSourceIndex, 0);
+}
+
 u16 HPI_SampleClock_GetSourceIndex(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2635,14 +2818,44 @@ u16 HPI_SampleClock_GetSourceIndex(
 	return (wError);
 }
 
-u16 HPI_SampleClock_SetSampleRate(
+u16 HPI_SampleClock_QueryLocalRate(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hClock,
+	const u32 dwIndex,
+	u32 *pdwRate
+)
+{
+	u16 err;
+	err = HPI_ControlQuery(phSubSys, hClock,
+		HPI_SAMPLECLOCK_LOCAL_SAMPLERATE, dwIndex, 0, pdwRate);
+
+	return err;
+}
+
+u16 HPI_SampleClock_SetLocalRate(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
 	u32 dwSampleRate
 )
 {
 	return HPI_ControlParamSet(phSubSys, hControl,
-		HPI_SAMPLECLOCK_SAMPLERATE, dwSampleRate, 0);
+		HPI_SAMPLECLOCK_LOCAL_SAMPLERATE, dwSampleRate, 0);
+}
+
+u16 HPI_SampleClock_GetLocalRate(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwSampleRate
+)
+{
+	u16 wError = 0;
+	u32 dwSampleRate = 0;
+	wError = HPI_ControlParam1Get(phSubSys, hControl,
+		HPI_SAMPLECLOCK_LOCAL_SAMPLERATE, &dwSampleRate);
+	if (!wError)
+		if (pdwSampleRate)
+			*pdwSampleRate = dwSampleRate;
+	return (wError);
 }
 
 u16 HPI_SampleClock_GetSampleRate(
@@ -2661,6 +2874,26 @@ u16 HPI_SampleClock_GetSampleRate(
 	return (wError);
 }
 
+u16 HPI_SampleClock_SetAuto(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 dwAuto
+)
+{
+	return HPI_ControlParamSet(phSubSys, hControl,
+		HPI_SAMPLECLOCK_AUTO, dwAuto, 0);
+}
+
+u16 HPI_SampleClock_GetAuto(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwAuto
+)
+{
+	return HPI_ControlParam1Get(phSubSys, hControl,
+		HPI_SAMPLECLOCK_AUTO, pdwAuto);
+}
+
 u16 HPI_ToneDetector_GetFrequency(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2832,6 +3065,22 @@ u16 HPI_SilenceDetector_GetThreshold(
 		HPI_SILENCEDETECTOR_THRESHOLD, 0, 0, (u32 *)Threshold, NULL);
 }
 
+u16 HPI_Tuner_QueryBand(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hTuner,
+	const u32 dwIndex,
+	u16 *pwBand
+)
+{
+	u32 qr;
+	u16 err;
+
+	err = HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_BAND,
+		dwIndex, 0, &qr);
+	*pwBand = (u16)qr;
+	return err;
+}
+
 u16 HPI_Tuner_SetBand(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2842,26 +3091,6 @@ u16 HPI_Tuner_SetBand(
 		wBand, 0);
 }
 
-u16 HPI_Tuner_SetGain(
-	struct hpi_hsubsys *phSubSys,
-	u32 hControl,
-	short nGain
-)
-{
-	return HPI_ControlParamSet(phSubSys, hControl, HPI_TUNER_GAIN,
-		nGain, 0);
-}
-
-u16 HPI_Tuner_SetFrequency(
-	struct hpi_hsubsys *phSubSys,
-	u32 hControl,
-	u32 wFreqInkHz
-)
-{
-	return HPI_ControlParamSet(phSubSys, hControl, HPI_TUNER_FREQ,
-		wFreqInkHz, 0);
-}
-
 u16 HPI_Tuner_GetBand(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2878,6 +3107,64 @@ u16 HPI_Tuner_GetBand(
 	return nError;
 }
 
+u16 HPI_Tuner_QueryFrequency(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hTuner,
+	const u32 dwIndex,
+	const u16 band,
+	u32 *pdwFreq
+)
+{
+	return HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_FREQ,
+		dwIndex, band, pdwFreq);
+}
+
+u16 HPI_Tuner_SetFrequency(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 wFreqInkHz
+)
+{
+	return HPI_ControlParamSet(phSubSys, hControl, HPI_TUNER_FREQ,
+		wFreqInkHz, 0);
+}
+
+u16 HPI_Tuner_GetFrequency(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pwFreqInkHz
+)
+{
+	return HPI_ControlParam1Get(phSubSys, hControl, HPI_TUNER_FREQ,
+		pwFreqInkHz);
+}
+
+u16 HPI_Tuner_QueryGain(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hTuner,
+	const u32 dwIndex,
+	u16 *pwGain
+)
+{
+	u32 qr;
+	u16 err;
+
+	err = HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_BAND,
+		dwIndex, 0, &qr);
+	*pwGain = (u16)qr;
+	return err;
+}
+
+u16 HPI_Tuner_SetGain(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	short nGain
+)
+{
+	return HPI_ControlParamSet(phSubSys, hControl, HPI_TUNER_GAIN,
+		nGain, 0);
+}
+
 u16 HPI_Tuner_GetGain(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2894,16 +3181,6 @@ u16 HPI_Tuner_GetGain(
 	return nError;
 }
 
-u16 HPI_Tuner_GetFrequency(
-	struct hpi_hsubsys *phSubSys,
-	u32 hControl,
-	u32 *pwFreqInkHz
-)
-{
-	return HPI_ControlParam1Get(phSubSys, hControl, HPI_TUNER_FREQ,
-		pwFreqInkHz);
-}
-
 u16 HPI_Tuner_GetRFLevel(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2940,6 +3217,92 @@ u16 HPI_Tuner_GetRawRFLevel(
 	return (hr.wError);
 }
 
+u16 HPI_Tuner_QueryDeemphasis(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hTuner,
+	const u32 dwIndex,
+	const u16 band,
+	u32 *pdwDeemphasis
+)
+{
+	return HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_DEEMPHASIS,
+		dwIndex, band, pdwDeemphasis);
+}
+
+u16 HPI_Tuner_SetDeemphasis(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 dwDeemphasis
+)
+{
+	return HPI_ControlParamSet(phSubSys, hControl, HPI_TUNER_DEEMPHASIS,
+		dwDeemphasis, 0);
+}
+
+u16 HPI_Tuner_GetDeemphasis(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwDeemphasis
+)
+{
+	return HPI_ControlParam1Get(phSubSys, hControl, HPI_TUNER_DEEMPHASIS,
+		pdwDeemphasis);
+}
+
+u16 HPI_Tuner_QueryProgram(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hTuner,
+	u32 *pbitmapProgram
+)
+{
+	return HPI_ControlQuery(phSubSys, hTuner, HPI_TUNER_PROGRAM,
+		0, 0, pbitmapProgram);
+}
+
+u16 HPI_Tuner_SetProgram(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 dwProgram
+)
+{
+	return HPI_ControlParamSet(phSubSys, hControl, HPI_TUNER_PROGRAM,
+		dwProgram, 0);
+}
+
+u16 HPI_Tuner_GetProgram(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwProgram
+)
+{
+	return HPI_ControlParam1Get(phSubSys, hControl, HPI_TUNER_PROGRAM,
+		pdwProgram);
+}
+
+u16 HPI_Tuner_GetHdRadioDspVersion(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	char *pszDspVersion,
+	const u32 dwStringSize
+)
+{
+	return HPI_Control_GetString(phSubSys,
+		hControl,
+		HPI_TUNER_HDRADIO_DSP_VERSION, pszDspVersion, dwStringSize);
+}
+
+u16 HPI_Tuner_GetHdRadioSdkVersion(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	char *pszSdkVersion,
+	const u32 dwStringSize
+)
+{
+	return HPI_Control_GetString(phSubSys,
+		hControl,
+		HPI_TUNER_HDRADIO_SDK_VERSION, pszSdkVersion, dwStringSize);
+}
+
 u16 HPI_Tuner_GetStatus(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -2986,6 +3349,16 @@ u16 HPI_Tuner_GetMode(
 		nMode, 0, pnValue, NULL);
 }
 
+u16 HPI_Tuner_GetHdRadioSignalQuality(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwQuality
+)
+{
+	return HPI_ControlParamGet(phSubSys, hControl,
+		HPI_TUNER_HDRADIO_SIGNAL_QUALITY, 0, 0, pdwQuality, NULL);
+}
+
 u16 HPI_Tuner_GetRDS(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -3006,6 +3379,80 @@ u16 HPI_Tuner_GetRDS(
 	return (hr.wError);
 }
 
+u16 HPI_PAD_GetChannelName(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	char *pszString,
+	const u32 dwDataLength
+)
+{
+	return HPI_Control_GetString(phSubSys,
+		hControl, HPI_PAD_CHANNEL_NAME, pszString, dwDataLength);
+}
+
+u16 HPI_PAD_GetArtist(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	char *pszString,
+	const u32 dwDataLength
+)
+{
+	return HPI_Control_GetString(phSubSys,
+		hControl, HPI_PAD_ARTIST, pszString, dwDataLength);
+}
+
+u16 HPI_PAD_GetTitle(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	char *pszString,
+	const u32 dwDataLength
+)
+{
+	return HPI_Control_GetString(phSubSys,
+		hControl, HPI_PAD_TITLE, pszString, dwDataLength);
+}
+
+u16 HPI_PAD_GetComment(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	char *pszString,
+	const u32 dwDataLength
+)
+{
+	return HPI_Control_GetString(phSubSys,
+		hControl, HPI_PAD_COMMENT, pszString, dwDataLength);
+}
+
+u16 HPI_PAD_GetProgramType(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwPTY
+)
+{
+	return HPI_ControlParamGet(phSubSys, hControl, HPI_PAD_PROGRAM_TYPE,
+		0, 0, pdwPTY, NULL);
+}
+
+u16 HPI_PAD_GetRdsPI(
+	struct hpi_hsubsys *phSubSys,
+	u32 hControl,
+	u32 *pdwPI
+)
+{
+	return HPI_ControlParamGet(phSubSys, hControl, HPI_PAD_PROGRAM_ID,
+		0, 0, pdwPI, NULL);
+}
+
+u16 HPI_Volume_QueryChannels(
+	const struct hpi_hsubsys *phSubSys,
+	const u32 hVolume,
+	u32 *pChannels
+)
+{
+	return HPI_ControlQuery(phSubSys, hVolume, HPI_VOLUME_NUM_CHANNELS,
+		0, 0, pChannels);
+}
+
 u16 HPI_VolumeSetGain(
 	struct hpi_hsubsys *phSubSys,
 	u32 hControl,
@@ -3085,14 +3532,13 @@ u16 HPI_VolumeAutoFadeProfile(
 	struct hpi_response hr;
 	HPI_InitMessage(&hm, HPI_OBJ_CONTROL, HPI_CONTROL_SET_STATE);
 	u32TOINDEXES(hControl, &hm.wAdapterIndex, &hm.u.c.wControlIndex);
+
 	memcpy(hm.u.c.anLogValue, anStopGain0_01dB,
 		sizeof(short) * HPI_MAX_CHANNELS);
 
-	if (wProfile < HPI_VOLUME_AUTOFADE)
-		wProfile = HPI_VOLUME_AUTOFADE;
-
-	hm.u.c.wAttribute = wProfile;
+	hm.u.c.wAttribute = HPI_VOLUME_AUTOFADE;
 	hm.u.c.dwParam1 = dwDurationMs;
+	hm.u.c.dwParam2 = wProfile;
 
 	HPI_Message(&hm, &hr);
 
@@ -3108,7 +3554,7 @@ u16 HPI_VolumeAutoFade(
 {
 	return HPI_VolumeAutoFadeProfile(phSubSys,
 		hControl,
-		anStopGain0_01dB, dwDurationMs, HPI_VOLUME_AUTOFADE);
+		anStopGain0_01dB, dwDurationMs, HPI_VOLUME_AUTOFADE_LINEAR);
 }
 
 u16 HPI_VoxSetThreshold(
@@ -3192,7 +3638,7 @@ u16 HPI_GpioReadBit(
 
 	HPI_Message(&hm, &hr);
 
-	*pwBitData = hr.u.l.wBitData;
+	*pwBitData = hr.u.l.wBitData[0];
 	return (hr.wError);
 }
 
@@ -3209,7 +3655,12 @@ u16 HPI_GpioReadAllBits(
 
 	HPI_Message(&hm, &hr);
 
-	*pwBitData = hr.u.l.wBitData;
+	if (pwBitData) {
+		pwBitData[0] = hr.u.l.wBitData[0];
+		pwBitData[1] = hr.u.l.wBitData[1];
+		pwBitData[2] = hr.u.l.wBitData[2];
+		pwBitData[3] = hr.u.l.wBitData[3];
+	}
 	return (hr.wError);
 }
 
@@ -3232,6 +3683,28 @@ u16 HPI_GpioWriteBit(
 	return (hr.wError);
 }
 
+u16 HPI_GpioWriteStatus(
+	struct hpi_hsubsys *phSubSys,
+	u32 hGpio,
+	u16 *pwBitData
+)
+{
+	struct hpi_message hm;
+	struct hpi_response hr;
+	HPI_InitMessage(&hm, HPI_OBJ_GPIO, HPI_GPIO_WRITE_STATUS);
+	u32TOINDEX(hGpio, &hm.wAdapterIndex);
+
+	HPI_Message(&hm, &hr);
+
+	if (pwBitData) {
+		pwBitData[0] = hr.u.l.wBitData[0];
+		pwBitData[1] = hr.u.l.wBitData[1];
+		pwBitData[2] = hr.u.l.wBitData[2];
+		pwBitData[3] = hr.u.l.wBitData[3];
+	}
+	return (hr.wError);
+}
+
 u16 HPI_AsyncEventOpen(
 	struct hpi_hsubsys *phSubSys,
 	u16 wAdapterIndex,
@@ -3632,14 +4105,14 @@ u16 HPI_FormatCreate(
 		break;
 	case HPI_FORMAT_PCM16_SIGNED:
 	case HPI_FORMAT_PCM16_BIGENDIAN:
-		Format.dwBitRate = (u32)wChannels *dwSampleRate * 2;
+		Format.dwBitRate = wChannels * dwSampleRate * 2;
 		break;
 	case HPI_FORMAT_PCM32_SIGNED:
 	case HPI_FORMAT_PCM32_FLOAT:
-		Format.dwBitRate = (u32)wChannels *dwSampleRate * 4;
+		Format.dwBitRate = wChannels * dwSampleRate * 4;
 		break;
 	case HPI_FORMAT_PCM8_UNSIGNED:
-		Format.dwBitRate = (u32)wChannels *dwSampleRate;
+		Format.dwBitRate = wChannels * dwSampleRate;
 		break;
 	default:
 		Format.dwBitRate = 0;
