@@ -1,4 +1,9 @@
+#include "config.h"
+#ifdef CONFIG_HAVE_DEPRECATED_CONFIG_H
+#include <linux/autoconf.h>
+#else
 #include <linux/config.h>
+#endif
 #include <linux/version.h>
 
 #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)
