@@ -1,6 +1,6 @@
 /*
  *  Utility to find module dependencies from Modules.dep
- *  Copyright (c) by Jaroslav Kysela <perex@suse.cz>
+ *  Copyright (c) by Jaroslav Kysela <perex@perex.cz>
  *		     Anders Semb Hermansen <ahermans@vf.telia.no>,
  *		     Martin Dahl <dahlm@vf.telia.no>,
  *
@@ -44,6 +44,7 @@ struct cond {
 	int type;
 	struct cond *next;
 	struct cond *stack_prev;
+	int state_prev;
 };
 
 struct sel {
@@ -53,6 +54,8 @@ struct sel {
 	struct sel *next;
 };
 
+enum { TYPE_TRISTATE, TYPE_BOOL, TYPE_INT };
+
 struct dep {
 	char *name;
 	// dependency part - conditions (chain)
@@ -62,9 +65,16 @@ struct dep {
 	// misc
 	struct dep *next;
 	// bool?
-	int is_bool;
+	int type;
+	// default value
+	int int_val;
+	int def_val;
 	// hitflag?
 	int hitflag;
+	int own_config;
+	int selectable;
+	int processed;
+	int pending;
 };
 
 // Prototypes
@@ -76,12 +86,14 @@ static void free_cond(struct cond *cond);
 static struct cond *create_cond(char *line);
 static struct dep *alloc_mem_for_dep(void);
 static struct dep *find_or_create_dep(char *line);
-static void add_dep(struct dep * dep, char *line, struct cond *template);
-static void add_select(struct dep * dep, char *line, struct cond *template);
+static void add_template(struct dep *dep, struct cond *template);
+static void add_dep(struct dep * dep, char *line);
+static void add_select(struct dep * dep, char *line);
 static char *get_word(char *line, char *word);
 static struct dep *find_dep(char *parent, char *depname);
 static void del_all_from_list(void);
 static int is_always_true(struct dep *dep);
+static int is_always_false(struct dep *dep);
 
 int main(int argc, char *argv[]);
 static void usage(char *programname);
@@ -99,6 +111,7 @@ static char *kernel_deps[] = {
 	"ISA",
 	"ISA_DMA_API",
 	"ISAPNP",
+	"PNP",
 	"EISA",
 	"PCI",
 	"SBUS",
@@ -107,76 +120,103 @@ static char *kernel_deps[] = {
 	"L3",
 	"USB",
 	"PCMCIA",
+	"SPI",
+	"I2C*",
 	/* architectures */
-	"ARM",
+	"ARM*",
 	"PARISC",
-	"SPARC32",
-	"SPARC64",
-	"PPC",
-	"PPC64",
-	"PPC_PMAC",
-	"X86_64",
-	"X86",
-	"MIPS",
-	"MIPS64",
+	"SPARC*",
+	"PPC*",
+	"X86*",
+	"MIPS*",
+	"SUPERH*",
 	"IA32_EMULATION",
+	"M68K",
+	"ALPHA*",
+	"BLACKFIN*",
 	/* architecture specific */
-	"ARCH_SA1100",
-	"ARCH_PXA",
+	"ARCH_*",
 	"X86_PC9800",
+	"SH_*",
+	"GSC",
+	"MACH_*",
+	"PXA_*",
+	"PS3*",
+	"SGI*",
+	"BOARD_*",
+	"ATMEL_*",
+	"AVR*",
+	"MPC*",
+	"MFD*",
+	"CPU_*",
+	"SOC_*",
 	/* other drivers */
 	"RTC",
+	"HPET",
+	"PARPORT",
+	"XILINX_VIRTEX",
 	"GAMEPORT",
-	"VIDEO_DEV",
+	"VIDEO_*",
 	"FW_LOADER",
-	/* some flags */
+	"PCSPKR_PLATFORM",
+	/* some flags/capabilities */
+	"PROC_FS",
+	"HIGH_RES_TIMERS",
+	"HAS_IOPORT",
 	"EXPERIMENTAL",
+	"BROKEN",
+	/* sound common */
+	"AC97_BUS",
+	/* workaround */
+	"SND_SOC_ALL_CODECS",
+	/* this is used only in i2c/cs8427.c, and we have a wrapper there */
+	"%BITREVERSE",
+	/* used in core/pcm_timer.c, a wrapper present */
+	"%GCD",
 	NULL
 };
 
 /* % -> always true */
-/* # -> define */
+/* ! -> always false */
+/* # -> menuconfig */
 static char *no_cards[] = {
-	"%#SOUND",
+	"%SOUND",
+	"%HAS_IOMEM",
 	"SOUND_PRIME",
 	"%SND",
-	"#SND_TIMER",
-	"#SND_HWDEP",
-	"#SND_RAWMIDI",
-	"#SND_PCM",
-	"SND_SEQUENCER",
-	"SND_MIXER_OSS",
-	"SND_PCM_OSS",
-	"SND_PCM_OSS_PLUGINS",
-	"SND_SEQUENCER_OSS",
-	"SND_OSSEMUL",
-	"SND_RTCTIMER",
-	"SND_HPET",
+	"!M68K",
+	/* options given as configure options */
 	"SND_DYNAMIC_MINORS",
 	"SND_DEBUG",
 	"SND_DEBUG_MEMORY",
-	"SND_DEBUG_DETECT",
+	"SND_DEBUG_VERBOSE",
 	"SND_VERBOSE_PROCFS",
 	"SND_VERBOSE_PRINTK",
-	"SND_BIT32_EMUL",
-	"#SND_OPL3_LIB",
-	"#SND_OPL4_LIB",
-	"#SND_VX_LIB",
-	"#SND_MPU401_UART",
-	"#SND_GUS_SYNTH",
-	"#SND_AC97_BUS",
-	"#SND_AC97_CODEC",
-	"#SND_PXA2XX_PCM",
-	"#SND_GENERIC_PM",
+	"SND_SEQUENCER",
+	"SND_OSSEMUL",
+	"SND_PCM_OSS",
+	"SND_MIXER_OSS",
+	"SND_PCM_OSS_PLUGINS",
+	"SND_SEQUENCER_OSS",
+	/* menuconfig with default yes (not appearing in configure options) */
+	/* these options don't create any modules by themselves but work only
+	 * as primary dependencies.
+	 */
+	"#SND_ISA",
+	"#SND_PCI",
+	"#SND_DRIVERS",
+	"#SND_USB",
+	"#SND_PCMCIA",
+	"#SND_ARM",
+	"#SND_MIPS",
+	"#SND_GSC",
+	"#SND_PPC",
+	"#SND_SUPERH",
+	"#SND_SPARC",
+	"#SND_SPI",
 	NULL
 };
 
-static const char *kernel26_opts[] = {
-	"SND_AC97_BUS",
-	"SND_PXA2XX_PCM",
-	NULL,
-};
-
 #define READ_STATE_NONE		0
 #define READ_STATE_CONFIG	1
 #define READ_STATE_MENU		2
@@ -195,9 +235,11 @@ static int read_file(const char *filename, struct cond **template)
 	
 	fullfile = malloc(strlen(basedir) + strlen(hiddendir) + 1 + strlen(filename) + 1);
 	sprintf(fullfile, "%s/%s", basedir, filename);
-	if ((err = read_file_1(fullfile, template)) < 0) {
-		free(fullfile);
-		return err;
+	if (access(fullfile, R_OK) == 0) {
+		if ((err = read_file_1(fullfile, template)) < 0) {
+			free(fullfile);
+			return err;
+		}
 	}
 	if (!strncmp(filename, "core/", 5))
 		sprintf(fullfile, "%s/acore/%s", hiddendir, filename + 5);
@@ -213,6 +255,25 @@ static int read_file(const char *filename, struct cond **template)
 	return 0;
 }
 
+static void parse_default_value(struct dep *dep, char *buffer)
+{
+	switch (*buffer) {
+	case 'y':
+	case 'm':
+		dep->def_val = 1;
+		break;
+	case 'n':
+		dep->def_val = 0;
+		break;
+	default:
+		fprintf(stderr, "can't handle 'default %s', translating as depends on\n",
+			buffer);
+		dep->def_val = 1;
+		add_dep(dep, buffer);
+		break;
+	}
+}
+
 static int read_file_1(const char *filename, struct cond **template)
 {
 	char *buffer, *newbuf;
@@ -260,68 +321,128 @@ static int read_file_1(const char *filename, struct cond **template)
 		/* ignore some keywords */
 		if (buffer[0] == '#')
 			continue;
-		if (!strncmp(buffer, "        ", 8)) {
-			buffer[0] = '\t';
-			for (idx = 8; idx <= strlen(buffer); idx++)
-				buffer[idx-7] = buffer[idx];
+		if (buffer[0] != '\t') {
+			for (c = 0; c <= 8; c++) {
+				if (c == 8) {
+					c--;
+					buffer[7] = '\t';
+				}
+				if (buffer[c] == '\t') {
+					buffer[0] = '\t';
+					memmove(buffer + 1, buffer + c + 1,
+						strlen(buffer) - c);
+					break;
+				}
+				if (buffer[c] != ' ')
+					break;
+			}
 		}
-		if (!strncmp(buffer, "endmenu", 7)) {
+		if (!strncmp(buffer, "endmenu", 7) ||
+		    !strncmp(buffer, "endif", 5)) {
 			struct cond *otemplate;
-		    	state = READ_STATE_NONE;
 		    	if (*template == NULL) {
 		    		fprintf(stderr, "Menu level error\n");
 		    		exit(EXIT_FAILURE);
 		    	}
 		    	otemplate = *template;
+		    	state = (*template)->state_prev;
 		    	*template = (*template)->stack_prev;
 		    	free_cond(otemplate);
 			continue;
-		}
-		if (!strncmp(buffer, "menu", 4)) {
-			struct cond *ntemplate;
+		} else if (!strncmp(buffer, "menuconfig ", 10)) {
+			state = READ_STATE_CONFIG;
+			dep = find_or_create_dep(buffer + 10);
+			if (dep == NULL) {
+				result = -ENOMEM;
+				goto __end;
+			}
+			add_template(dep, *template);
+			dep->own_config = 1;
+			continue;
+		} else if (!strncmp(buffer, "if ", 3)) {
+			struct cond *ntemp;
+			ntemp = create_cond(buffer + 3);
+			ntemp->stack_prev = *template;
+			ntemp->state_prev = state;
+			*template= ntemp;
 			state = READ_STATE_MENU;
+			continue;
+		} else if (!strncmp(buffer, "menu", 4)) {
+			struct cond *ntemplate;
 			strcpy(buffer, "EMPTY");
 			ntemplate = create_cond(buffer);
 			ntemplate->stack_prev = *template;
+			ntemplate->state_prev = state;
 			*template = ntemplate;
+			state = READ_STATE_MENU;
 			continue;
-		}
-		if (!strncmp(buffer, "config ", 7)) {
+		} else if (!strncmp(buffer, "config ", 7)) {
 			state = READ_STATE_CONFIG;
 			dep = find_or_create_dep(buffer + 7);
 			if (dep == NULL) {
 				result = -ENOMEM;
 				goto __end;
 			}
-		}
-		if (!strncmp(buffer, "source ", 7)) {
+			add_template(dep, *template);
+			dep->own_config = 1;
+			continue;
+		} else if (!strncmp(buffer, "source ", 7)) {
 			state = READ_STATE_NONE;
 			result = include_file(buffer + 7, template);
 			if (result < 0)
 				goto __end;
-		}
-		if (!strncmp(buffer, "comment", 7)) {
+			continue;
+		} else if (!strncmp(buffer, "comment", 7)) {
 			state = READ_STATE_COMMENT;
+			continue;
 		}
 		switch (state) {
 		case READ_STATE_CONFIG:
-			if (!strncmp(buffer, "\ttristate", 9))
-				dep->is_bool = 0;
-			else if (!strncmp(buffer, "\tbool", 5))
-				dep->is_bool = 1;
-			else if (!strncmp(buffer, "\tdepends on ", 12))
-				add_dep(dep, buffer + 12, *template);
+			if (!strncmp(buffer, "\ttristate", 9)) {
+				dep->type = TYPE_TRISTATE;
+				if (buffer[9])
+					dep->selectable = 1;
+			} else if (!strncmp(buffer, "\tbool", 5)) {
+				dep->type = TYPE_BOOL;
+				if (buffer[5])
+					dep->selectable = 1;
+			} else if (!strncmp(buffer, "\tdef_bool ", 10)) {
+				dep->type = TYPE_BOOL;
+				parse_default_value(dep, buffer + 10);
+			} else if (!strncmp(buffer, "\tdef_tristate ", 14)) {
+				dep->type = TYPE_TRISTATE;
+				parse_default_value(dep, buffer + 14);
+			} else if (!strncmp(buffer, "\tint ", 5)) {
+				dep->type = TYPE_INT;
+				/*dep->selectable = 1;*/ /* exception */
+			} else if (!strncmp(buffer, "\tdepends on ", 12))
+				add_dep(dep, buffer + 12);
+			else if (!strncmp(buffer, "\tdepends ", 9))
+				add_dep(dep, buffer + 9);
 			else if (!strncmp(buffer, "\tselect ", 8))
-				add_select(dep, buffer + 8, *template);
+				add_select(dep, buffer + 8);
+			if (!strncmp(buffer, "\tdefault ", 9)) {
+				if (dep->type == TYPE_INT) {
+					char *p = buffer + 9;
+					for (; *p && !isdigit(*p); p++)
+						;
+					dep->int_val = strtol(p, NULL, 0);
+				} else {
+					parse_default_value(dep, buffer + 9);
+				}
+			}
 			continue;
 		case READ_STATE_MENU:
-			if (!strncmp(buffer, "\tdepends on ", 12)) {
+			if (!strncmp(buffer, "\tdepends ", 9)) {
 				struct cond *ntemplate;
 				if (strcmp((*template)->name, "EMPTY")) {
 					fprintf(stderr, "Menu consistency error\n");
 					exit(EXIT_FAILURE);
 				}
-				ntemplate = create_cond(buffer + 12);
+				if (! strncmp(buffer + 9, "on ", 3))
+					ntemplate = create_cond(buffer + 12);
+				else
+					ntemplate = create_cond(buffer + 9);
 				free((*template)->name);
 				(*template)->name = ntemplate->name;
 				(*template)->dep = ntemplate->dep;
@@ -367,6 +488,9 @@ static struct cond * create_cond(char *line)
 	char *word = NULL;
 	int i;
 
+	if (!line)
+		return NULL;
+
 	word = malloc(strlen(line) + 1);
 	if (word == NULL)
 		nomem();
@@ -383,6 +507,12 @@ static struct cond * create_cond(char *line)
 			for (i = 1; i < strlen(word) + 1; i++)
 				word[i - 1] = word[i];
 			cond->left++;
+			if (!word[0]) {
+				if (!get_word(line, word)) {
+					fprintf(stderr, "Unbalanced open-parenthesis\n");
+					exit(EXIT_FAILURE);
+				}
+			}
 		}
 		/* hack for !XXX */
 		if (word[0] == '!' && isascii(word[1])) {
@@ -401,6 +531,17 @@ static struct cond * create_cond(char *line)
 					word[i] = '\0';
 					break;
 				}
+				if (word[i] == '=') {
+					if (word[i + 1] == 'n' && !cond->not) {
+						cond->not = 1;
+						word[i] = '\0';
+						break;
+					}
+					fprintf(stderr, "can't handle word %s properly, supposing it's OK\n",
+						word);
+					word[i] = '\0';
+					break;
+				}
 				fprintf(stderr, "Unknown suffix '%s'\n", word + i);
 				exit(EXIT_FAILURE);
 			}
@@ -410,15 +551,25 @@ static struct cond * create_cond(char *line)
 			nomem();
 		if (strcmp(cond->name, "EMPTY"))
 			find_or_create_dep(word);
-		if (get_word(line, word)) {
+		while (get_word(line, word)) {
 			if (!strcmp(word, "&&"))
 				cond->type = COND_AND;
 			else if (!strcmp(word, "||"))
 				cond->type = COND_OR;
-			else {
+			else if (!strcmp(word, "!=")) {
+				get_word(line, word);
+				continue;
+			} else if (!strcmp(word, "!=n"))
+				continue;
+			else if (*word == ')') {
+				for (i = 0; word[i]; i++)
+					if (word[i] == ')')
+						cond->right++;
+			} else {
 				fprintf(stderr, "Wrong condition %s\n", word);
 				exit(EXIT_FAILURE);
 			}
+			break;
 		}
 	}
 	free(word);
@@ -457,6 +608,7 @@ static struct dep * find_or_create_dep(char *line)
 	new_dep->name = strdup(word);	// Fill in name of dependency
 	if (new_dep->name == NULL)
 		nomem();
+	//fprintf(stderr, "xxx created new dep %s\n", word);
 	free(word);
 	return new_dep;
 }
@@ -516,18 +668,44 @@ static struct cond *join_cond(struct cond *cond1, struct cond *cond2)
 	while (cond1->next)
 		cond1 = cond1->next;
 	cond1->next = cond2;
+	cond2->left += 1;
+	while (cond2->next)
+		cond2 = cond2->next;
+	cond2->right += 1;
 	return orig;
 }
 
 // Add a new dependency to the current one
-static void add_dep(struct dep * dep, char *line, struct cond *template)
+static void add_template(struct dep *dep, struct cond *template)
 {
 	template = duplicate_cond(template);
-	dep->cond = join_cond(template, create_cond(line));
+	dep->cond = join_cond(dep->cond, template);
+}
+
+static void add_dep(struct dep * dep, char *line)
+{
+	dep->cond = join_cond(dep->cond, create_cond(line));
+}
+
+/* Is the selected item ALSA-specific? */
+static int is_alsa_item(const char *word)
+{
+	static const char *known_items[] = {
+		"AC97_BUS",
+		NULL
+	};
+	const char **p;
+
+	if (!strncmp(word, "SND_", 4))
+		return 1;
+	for (p = known_items; *p; p++)
+		if (!strcmp(word, *p))
+			return 1;
+	return 0;
 }
 
 // Add a new forced (selected) dependency to the current one
-static void add_select(struct dep * dep, char *line, struct cond *template)
+static void add_select(struct dep * dep, char *line)
 {
 	char *word = NULL;
 	struct sel *sel, *nsel;
@@ -536,8 +714,8 @@ static void add_select(struct dep * dep, char *line, struct cond *template)
 	if (word == NULL)
 		nomem();
 	get_word(line, word);
-	if (strncmp(word, "SND_", 4)) {
-		add_dep(dep, word, template);
+	if (!is_alsa_item(word)) {
+		add_dep(dep, word);
 		free(word);
 		return;
 	}
@@ -556,6 +734,7 @@ static void add_select(struct dep * dep, char *line, struct cond *template)
 			sel = sel->next;
 		sel->next = nsel;
 	}
+	//fprintf(stderr, "*** add %s -> %s\n", word, dep->name);
 	free(word);
 }
 
@@ -611,6 +790,23 @@ static char *get_word(char *line, char *word)
 	return word;
 }
 
+static int is_kernel_deps(const char *name)
+{
+	char **p;
+	char *q;
+
+	for (p = kernel_deps; *p; p++) {
+		if (!strcmp(*p, name))
+			return 1;
+		q = strchr(*p, '*');
+		if (!q)
+			continue;
+		if (!strncmp(*p, name, q - *p))
+			return 1;
+	}
+	return 0;
+}
+
 // Find the dependency named "depname"
 static struct dep *find_dep(char *parent, char *depname)
 {
@@ -618,15 +814,13 @@ static struct dep *find_dep(char *parent, char *depname)
 	int idx;
 
 	while (temp_dep) {
-		// fprintf(stderr, "depname = '%s', name = '%s'\n", depname, temp_dep->name);
+		//fprintf(stderr, "depname = '%s', name = '%s'\n", depname, temp_dep->name);
 		if (!strcmp(depname, temp_dep->name))
 			return temp_dep;
 		temp_dep = temp_dep->next;
 	}
-	for (idx = 0; kernel_deps[idx]; idx++) {
-		if (!strcmp(kernel_deps[idx], depname))
-			return NULL;
-	}
+	if (is_kernel_deps(depname))
+		return NULL;
 	if (strcmp(parent, "<root>"))
 		fprintf(stderr, "Warning: Unsatisfied dep for %s: %s\n", parent, depname);
 	return NULL;
@@ -666,14 +860,20 @@ static void optimize_dep(struct dep * parent)
 				cond->left = 0;
 			}
 			if (cond->next &&
-			    !strcmp(cond->name, cond->next->name) &&
-			    ((cond->left == cond->right &&
-			      cond->next->left == cond->next->right) ||
-			     (cond->left == cond->right + 1 &&
-			      cond->next->left + 1 == cond->next->right)) &&
-			      cond->left == cond->next->right)
-			      	goto __remove;
-			if (!is_always_true(cond->dep))
+			    !strcmp(cond->name, cond->next->name)) {
+			    	if (cond->left == cond->right &&
+			            cond->next->left == cond->next->right &&
+			            cond->left == cond->next->right)
+			            	goto __remove;
+				if (cond->left == cond->right + 1 &&
+				    cond->next->left + 1 == cond->next->right &&
+				    cond->left == cond->next->right) {
+					cond->next->right--;
+				      	goto __remove;
+			      	}
+			}
+			if (!(is_always_true(cond->dep) ||
+			      (is_always_false(cond->dep) && cond->not)))
 				goto __next;
 			if (cond->left == cond->right &&
 			    (cond->next == NULL || cond->type == COND_AND) &&
@@ -779,68 +979,53 @@ static int is_toplevel(struct dep *dep)
 	
 	if (dep == NULL)
 		return 0;
-	for (idx = 0; kernel_deps[idx]; idx++) {
-		if (!strcmp(kernel_deps[idx], dep->name))
-			return 0;
-	}
-	for (idx = 0; no_cards[idx]; idx++) {
-		str = no_cards[idx];
-		if (*str == '%' || *str == '#')
-			str++;
-		if (!strcmp(str, dep->name))
-			return 0;
-	}
-	return 1;
-}
-
-// whether to output AC_DEFINE() for this variable
-static int output_ac_define(struct dep *dep)
-{
-	int idx;
-	char *str;
-	
-	if (dep == NULL)
+	if (is_kernel_deps(dep->name))
 		return 0;
-	for (idx = 0; kernel_deps[idx]; idx++) {
-		if (!strcmp(kernel_deps[idx], dep->name))
-			return 0;
-	}
 	for (idx = 0; no_cards[idx]; idx++) {
-		int def = 0;
 		str = no_cards[idx];
 		if (*str == '%')
 			str++;
-		if (*str == '#') {
-			def = 1;
-			str++;
-		}
 		if (!strcmp(str, dep->name))
-			return def;
+			return 0;
 	}
 	return 1;
 }
 
-// is CONFIG_ variable is always true
-static int is_always_true(struct dep *dep)
+static int check_in_no_cards(struct dep *dep, char flag)
 {
-	int idx;
-	char *str;
+	char **p;
 
-	if (dep == NULL)
+	if (!dep)
 		return 0;
-	for (idx = 0; no_cards[idx]; idx++) {
-		str = no_cards[idx];
-		if (*str != '%')
+	for (p = no_cards; *p; p++) {
+		if (**p != flag)
 			continue;
-		str++;
-		if (!strcmp(str, dep->name))
+		if (!strcmp(*p + 1, dep->name))
 			return 1;
 	}
 	return 0;
 }
+// is CONFIG_ variable is always true
+static int is_always_true(struct dep *dep)
+{
+	return check_in_no_cards(dep, '%');
+}
+
+// is CONFIG_ variable is always false
+static int is_always_false(struct dep *dep)
+{
+	return check_in_no_cards(dep, '!');
+}
+
+/* is menuconfig that doesn't create modules and as default true */
+static int is_menu_default_yes(struct dep *dep)
+{
+	return check_in_no_cards(dep, '#');
+}
 
 // Print out ALL deps for firstdep (Cards, Deps)
-static void output_card_list(struct dep *firstdep, int space, int size, int tristate)
+static void output_card_list(struct dep *firstdep, int space, int size,
+			     int elem_type)
 {
 	struct dep *temp_dep=firstdep;
 	char *card_name;
@@ -849,14 +1034,14 @@ static void output_card_list(struct dep *firstdep, int space, int size, int tris
 	printf("  [");
 	for (idx = 0; idx < space; idx++)
 		printf(" ");
-	while(temp_dep) {
-		if (!is_toplevel(temp_dep))
-			goto __skip;
-		if (tristate && temp_dep->is_bool)
-			goto __skip;
-		if (!tristate && !temp_dep->is_bool)
-			goto __skip;
-		card_name=get_card_name(temp_dep->name);
+	for (; temp_dep; temp_dep = temp_dep->next) {
+		if (!temp_dep->selectable || !is_toplevel(temp_dep))
+			continue;
+		if (is_menu_default_yes(temp_dep))
+			continue;
+		if (temp_dep->type != elem_type)
+			continue;
+		card_name = get_card_name(temp_dep->name);
 		if (card_name) {
 			if (!first) {
 				printf(", ");
@@ -874,8 +1059,6 @@ static void output_card_list(struct dep *firstdep, int space, int size, int tris
 			tmp_size += strlen(card_name);
 			free(card_name);
 		}
-	      __skip:
-		temp_dep=temp_dep->next;
 	}
 }
 
@@ -892,21 +1075,103 @@ static void sel_remove_hitflags(void)
 	}
 }
 
-// is 2.6-only config option
-static int is_kernel26(const char *name)
+struct kconfig_verdep {
+	const char *name;
+	const char *version;
+	struct kconfig_verdep *next;
+};
+
+static struct kconfig_verdep *version_deps;
+
+static char *get_token(char **pp, int need_space)
 {
-	const char **p;
-	for (p = kernel26_opts; *p; p++)
-		if (! strcmp(name, *p))
-			return 1;
+	char *token;
+	char *p = *pp;
+
+	for (; isspace(*p); p++)
+		if (!*p)
+			return NULL;
+	if (*p == '%')
+		return NULL;
+	token = p;
+	for (; !isspace(*p); p++)
+		if (!*p) {
+			if (need_space)
+				return NULL;
+			*pp = p;
+			return token;
+		}
+	*p++ = 0;
+	*pp = p;
+	return token;
+}
+
+static int read_version_deps(const char *fname)
+{
+	FILE *fp;
+	char buf[128], *p, *name, *val;
+	struct kconfig_verdep *ver;
+
+	fp = fopen(fname, "r");
+	if (!fp) {
+		fprintf(stderr, "cannot open %s\n", fname);
+		return -ENODEV;
+	}
+	while (fgets(buf, sizeof(buf), fp)) {
+		p = buf;
+		name = get_token(&p, 1);
+		if (!name)
+			continue;
+		val = get_token(&p, 0);
+		if (!val)
+			continue;
+		ver = malloc(sizeof(*ver));
+		if (!ver)
+			return -ENOMEM;
+		ver->name = strdup(name);
+		ver->version = strdup(val);
+		ver->next = version_deps;
+		version_deps = ver;
+	}
+	fclose(fp);
 	return 0;
 }
 
+static const char *get_version_dep(const char *name)
+{
+	const struct kconfig_verdep *p;
+	for (p = version_deps; p; p = p->next)
+		if (! strcmp(name, p->name))
+			return p->version;
+	return NULL;
+}
+
+static void print_version_dep(const char *ver)
+{
+	const char *p;
+	printf("test \"$kversion.$kpatchlevel\" = \"");
+	for (p = ver; *p; p++) {
+		putchar(*p);
+		if (*p == '.') {
+			p++;
+			break;
+		}
+	}
+	for (; *p && *p != '.'; p++)
+		putchar(*p);
+	putchar('"');
+	if (*p) {
+		printf(" -a $ksublevel -ge ");
+		for (p++; *p && *p != '.'; p++)
+			putchar(*p);
+	}
+}
 
 static void sel_print_acinclude(struct sel *sel)
 {
 	struct dep * dep;
 	struct sel * nsel;
+	const char *ver;
 
 	dep = sel->dep;
 	if (dep == NULL)
@@ -921,26 +1186,184 @@ static void sel_print_acinclude(struct sel *sel)
 		if (!nsel->dep->hitflag) {
 			nsel->dep->hitflag = 1;
 			printf("      ");
-			if (is_kernel26(nsel->name))
-				printf("test $kpatchlevel = 6 && ");
+			ver = get_version_dep(nsel->name);
+			if (ver) {
+				print_version_dep(ver);
+				printf(" && ");
+			}
 			printf("CONFIG_%s=\"%c\"\n", nsel->name,
-				(nsel->dep && nsel->dep->is_bool) ? 'y' : 'm');
+				(nsel->dep && nsel->dep->type == TYPE_BOOL) ? 'y' : 'm');
 		}
 	}
 }
 
+				    
+static int to_be_pending(struct dep *dep)
+{
+	if (dep->pending)
+		return 1;
+	return (!dep->processed && !dep->selectable &&
+		!is_kernel_deps(dep->name) &&
+		!is_menu_default_yes(dep) &&
+		!is_always_true(dep) &&
+		!is_always_false(dep));
+}
+
+static void process_dep_acinclude(struct dep *tempdep, int slave,
+				  int process_pending)
+{
+	struct cond *cond, *cond_prev;
+	int put_topif, put_define, put_if;
+	int j;
+	char *text;
+	const char *ver;
+	struct sel *sel;
+
+	if (!tempdep->selectable) {
+		if (tempdep->type != TYPE_INT && !slave) {
+			if (!tempdep->cond || !tempdep->def_val)
+				return;
+		}
+	}
+	if (!is_toplevel(tempdep) || tempdep->processed)
+		return;
+#if 0 /* debug */
+	if (tempdep->cond) {
+		fprintf(stderr, "xxx DEP %s\n", tempdep->name);
+		for (cond = tempdep->cond; cond; cond = cond->next)
+			fprintf(stderr, "xxx   -> %s\n", cond->name);
+	}
+#endif
+	if (tempdep->processed)
+		return;
+
+	if (tempdep->cond) {
+		for (cond = tempdep->cond; cond; cond = cond->next) {
+			struct dep *dep;
+			dep = find_dep("<root>", cond->name);
+			if (dep) {
+				/* depends on a reverse-selected item? */
+				if (!process_pending && to_be_pending(dep)) {
+#if 0
+					fprintf(stderr, "pending %s (dep %s)\n",
+						tempdep->name, dep->name);
+#endif
+					tempdep->pending = 1;
+					return; /* pending */
+				}
+				process_dep_acinclude(dep, 1, process_pending);
+			}
+		}
+	}
+	if (tempdep->processed)
+		return;
+
+	put_topif = 0;
+	put_define = 0;
+	if (tempdep->selectable && !is_always_true(tempdep)) {
+		text = get_card_name(tempdep->name);
+		if (!text)
+			return;
+		if (!is_menu_default_yes(tempdep)) {
+			if (tempdep->type == TYPE_BOOL)
+				printf("  if alsa_check_kconfig_option \"%s\"; then\n", text);
+			else
+				printf("  if alsa_check_kconfig_card \"%s\"; then\n", text);
+			put_topif = 1;
+		}
+		put_define = 1;
+		free(text);
+	} else if (tempdep->type == TYPE_INT) {
+		put_define = 1;
+	} else if (tempdep->sel) {
+		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
+		if (!text)
+			return;
+		printf("  if test \"$CONFIG_%s\" = \"m\" -o \"$CONFIG_%s\" = \"y\"; then\n", text, text);
+		free(text);
+		put_topif = 1;
+	} else if (tempdep->cond && tempdep->def_val) {
+		/* "def_bool yes" and depends on ... */
+		put_define = 1;
+	} else
+		return;
+	put_if = 0;
+	for (cond = tempdep->cond, cond_prev = NULL; cond;
+	     cond_prev = cond, cond = cond->next) {
+		if (!put_if)
+			printf("    if ");
+		else {
+			printf(cond_prev->type == COND_AND ? " &&" : " ||");
+			printf("\n      ");
+		}
+		if (cond->not)
+			printf(" ! ");
+		for (j = 0; j < cond->left; j++)
+			printf("( ");
+		printf("( test \"$CONFIG_%s\" = \"y\" -o \"$CONFIG_%s\" = \"m\" )", cond->name, cond->name);
+		for (j = 0; j < cond->right; j++)
+			printf(" )");
+		put_if = 1;
+	}
+	text = convert_to_config_uppercase("", tempdep->name);
+	ver = get_version_dep(text);
+	if (ver) {
+		if (!put_if)
+			printf("    if ");
+		else {
+			printf(cond_prev->type == COND_AND ? " &&" : " ||");
+			printf("\n      ");
+		}
+		printf("( ");
+		print_version_dep(ver);
+		printf(" )");
+		put_if = 1;
+	}
+	free(text);
+	if (put_if)
+		printf("; then\n");
+
+	sel_remove_hitflags();
+	for (sel = tempdep->sel; sel; sel = sel->next)
+		sel_print_acinclude(sel);			
+	for (sel = tempdep->sel; sel; sel = sel->next) {
+		if (sel->dep->type == TYPE_INT)
+			continue;
+		if (is_always_true(sel->dep))
+			continue;
+		printf("      ");
+		ver = get_version_dep(sel->name);
+		if (ver) {
+			print_version_dep(ver);
+			printf(" && ");
+		}
+		printf("CONFIG_%s=\"%c\"\n", sel->name,
+		       (sel->dep && sel->dep->type == TYPE_BOOL) ? 'y' : 'm');
+	}
+	if (put_define) {
+		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
+		if (tempdep->type == TYPE_INT)
+			printf("      %s=\"%d\"\n", text, tempdep->int_val);
+		else
+			printf("      %s=\"%c\"\n", text, tempdep->type == TYPE_BOOL ? 'y' : 'm');
+		free(text);
+	}
+	if (put_if)
+		printf("    fi\n");
+	if (put_topif)
+		printf("  fi\n");
+	tempdep->processed = 1;
+}
+
 // Output in acinlude.m4
 static void output_acinclude(void)
 {
 	struct dep *tempdep;
 	char *text;
-	struct cond *cond, *cond_prev;
-	struct sel *sel;
-	int j;
 	
 	printf("dnl ALSA soundcard configuration\n");
 	printf("dnl Find out which cards to compile driver for\n");
-	printf("dnl Copyright (c) by Jaroslav Kysela <perex@suse.cz>,\n");
+	printf("dnl Copyright (c) by Jaroslav Kysela <perex@perex.cz>,\n");
 	printf("dnl                  Anders Semb Hermansen <ahermans@vf.telia.no>\n\n");
 
 	printf("AC_DEFUN([ALSA_TOPLEVEL_INIT], [\n");
@@ -953,157 +1376,82 @@ static void output_acinclude(void)
 
 	printf("AC_DEFUN([ALSA_TOPLEVEL_SELECT], [\n");
 	printf("dnl Check for which cards to compile driver for...\n");
-	printf("AC_MSG_CHECKING(for which soundcards to compile driver for)\n");
+	printf("AC_MSG_CHECKING(for cards to compile driver for)\n");
 	printf("AC_ARG_WITH(cards,\n"
-	       "  [  --with-cards=<list>     compile driver for cards in <list>; ]\n"
+	       "  [  --with-cards=<list> compile driver for cards and options in <list>; ]\n"
 	       "  [                        cards may be separated with commas; ]\n"
 	       "  [                        'all' compiles all drivers; ]\n"
 	       "  [                        Possible cards are: ]\n");
-	output_card_list(all_deps, 26, 50, 1);
+	output_card_list(all_deps, 26, 50, TYPE_TRISTATE);
 	printf(" ],\n");
 	printf("  cards=\"$withval\", cards=\"all\")\n");
 	printf("SELECTED_CARDS=`echo $cards | sed 's/,/ /g'`\n");
-	/* check cards */
-	printf("for card in $SELECTED_CARDS; do\n"
-	       "  probed=\n");
-	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
-		int put_if;
-		if (!is_toplevel(tempdep) || tempdep->is_bool)
-			continue;
-		text = get_card_name(tempdep->name);
-		if (! text)
-			continue;
-		printf("  if test \"$card\" = \"all\" -o \"$card\" = \"%s\"; then\n", text);
-		free(text);
-		put_if = 0;
-		for (cond = tempdep->cond, cond_prev = NULL; cond; cond = cond->next) {
-			if (!put_if)
-				printf("    if ");
-			else {
-				printf(cond_prev->type == COND_AND ? " &&" : " ||");
-				printf("\n      ");
-			}
-			if (cond->not)
-				printf(" ! ");
-			for (j = 0; j < cond->left; j++)
-				printf("( ");
-			printf("( test \"$CONFIG_%s\" = \"y\" -o \"$CONFIG_%s\" = \"m\" )", cond->name, cond->name);
-			for (j = 0; j < cond->right; j++)
-				printf(" )");
-			put_if = 1;
-			cond_prev = cond;
-		}
-		if (put_if)
-			printf("; then\n");
-
-		sel_remove_hitflags();
-		for (sel = tempdep->sel; sel; sel = sel->next)
-			sel_print_acinclude(sel);			
-		for (sel = tempdep->sel; sel; sel = sel->next) {
-			if (is_always_true(sel->dep))
-				continue;
-			printf("      ");
-			if (is_kernel26(sel->name))
-				printf("test $kpatchlevel = 6 && ");
-			printf("CONFIG_%s=\"%c\"\n", sel->name,
-			       (sel->dep && sel->dep->is_bool) ? 'y' : 'm');
-		}
-		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
-		printf("      %s=\"%c\"\n", text, tempdep->is_bool ? 'y' : 'm');
-		free(text);
-		printf("      probed=1\n");
-		if (put_if)
-			printf("    elif test -z \"$probed\"; then\n"
-			       "      probed=0\n"
-			       "    fi\n");
-		printf("  fi\n");
-	}
-	printf("  if test -z \"$probed\"; then\n"
-	       "    AC_MSG_ERROR(Unknown soundcard $card)\n"
-	       "  elif test \"$probed\" = \"0\"; then\n"
-	       "    AC_MSG_ERROR(Unsupported soundcard $card)\n"
-	       "  fi\n"
-	       "done\n\n");
-	/* options */
+	printf("AC_MSG_RESULT($SELECTED_CARDS)\n");
+	printf("AC_MSG_CHECKING(for additonal options to compile driver for)\n");
 	printf("AC_ARG_WITH(card_options,\n"
 	       "  [  --with-card-options=<list> enable driver options in <list>; ]\n"
 	       "  [                        options may be separated with commas; ]\n"
 	       "  [                        'all' enables all options; ]\n"
 	       "  [                        Possible options are: ]\n");
-	output_card_list(all_deps, 26, 50, 0);
+	output_card_list(all_deps, 26, 50, TYPE_BOOL);
 	printf(" ],\n");
 	printf("  cards=\"$withval\", cards=\"all\")\n");
 	printf("SELECTED_OPTIONS=`echo $cards | sed 's/,/ /g'`\n");
-	/* check cards */
-	printf("for card in $SELECTED_OPTIONS; do\n"
-	       "  probed=\n");
-	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
-		int put_if;
-		if (!is_toplevel(tempdep) || !tempdep->is_bool)
-			continue;
-		text = get_card_name(tempdep->name);
-		if (! text)
-			continue;
-		printf("  if test \"$card\" = \"all\" -o \"$card\" = \"%s\"; then\n", text);
-		free(text);
-		put_if = 0;
-		for (cond = tempdep->cond, cond_prev = NULL; cond; cond = cond->next) {
-			if (!put_if)
-				printf("    if ");
-			else {
-				printf(cond_prev->type == COND_AND ? " &&" : " ||");
-				printf("\n      ");
-			}
-			if (cond->not)
-				printf(" ! ");
-			for (j = 0; j < cond->left; j++)
-				printf("( ");
-			printf("( test \"$CONFIG_%s\" = \"y\" -o \"$CONFIG_%s\" = \"m\" )", cond->name, cond->name);
-			for (j = 0; j < cond->right; j++)
-				printf(" )");
-			put_if = 1;
-			cond_prev = cond;
-		}
-		if (put_if)
-			printf("; then\n");
+	printf("AC_MSG_RESULT($SELECTED_OPTIONS)\n");
+	printf("])\n");
 
-		sel_remove_hitflags();
-		for (sel = tempdep->sel; sel; sel = sel->next)
-			sel_print_acinclude(sel);			
-		for (sel = tempdep->sel; sel; sel = sel->next) {
-			if (is_always_true(sel->dep))
-				continue;
-			printf("      ");
-			if (is_kernel26(sel->name))
-				printf("test $kpatchlevel = 6 && ");
-			printf("CONFIG_%s=\"%c\"\n", sel->name,
-			       (sel->dep && sel->dep->is_bool) ? 'y' : 'm');
-		}
-		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
-		printf("      %s=\"y\"\n", text);
-		free(text);
-		printf("      probed=1\n");
-		if (put_if)
-			printf("    elif test -z \"$probed\"; then\n"
-			       "      probed=0\n"
-			       "    fi\n");
-		printf("  fi\n");
+	/* parse down the tree */
+	printf("AC_DEFUN([ALSA_PARSE_KCONFIG], [\n");
+
+	/* two helper functions */
+	printf("alsa_check_kconfig_card () {\n"
+	       "  local pat=${1}\n"
+	       "  for i in $SELECTED_CARDS; do\n"
+	       "    case \"$i\" in\n"
+	       "    $pat=n)\n"
+	       "      return 1;;\n"
+	       "    all|$pat|$pat=*)\n"
+	       "      return 0;;\n"
+	       "    esac\n"
+	       "  done\n"
+	       "  return 1\n"
+	       "}\n"
+	       "alsa_check_kconfig_option () {\n"
+	       "  local pat=${1}\n"
+	       "  for i in $SELECTED_OPTIONS; do\n"
+	       "    case \"$i\" in\n"
+	       "    $pat=n)\n"
+	       "      return 1;;\n"
+	       "    all|$pat|$pat=*)\n"
+	       "      return 0;;\n"
+	       "    esac\n"
+	       "  done\n"
+	       "  return 1\n"
+	       "}\n");
+
+	/* default SND=m */
+	printf("  CONFIG_SND=\"m\"\n");
+
+	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
+		process_dep_acinclude(tempdep, 0, 0);
 	}
-	printf("  if test -z \"$probed\"; then\n"
-	       "    AC_MSG_ERROR(Unknown option $card)\n"
-	       "  elif test \"$probed\" = \"0\" -a \"$card\" != \"all\"; then\n"
-	       "    AC_MSG_ERROR(Unsupported option $card)\n"
-	       "  fi\n"
-	       "done\n\n");
-
-	printf("AC_MSG_RESULT($SELECTED_CARDS)\n\n");
-	printf("CONFIG_SND=\"m\"\n");
+	/* process pending items */
 	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
-		if (!output_ac_define(tempdep))
+		process_dep_acinclude(tempdep, 0, 1);
+	}
+
+	printf("])\n\n");
+
+	printf("AC_DEFUN([ALSA_TOPLEVEL_DEFINES], [\n");
+	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
+		if (!tempdep->own_config)
 			continue;
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
 		printf("if test -n \"$%s\"; then\n", text);
-		if (tempdep->is_bool)
+		if (tempdep->type == TYPE_INT) {
+			printf("  AC_DEFINE_UNQUOTED([%s], [%d])\n",
+			       text, tempdep->int_val);
+		} else if (tempdep->type == TYPE_BOOL)
 			printf("  AC_DEFINE(%s)\n", text);
 		else
 			printf("  AC_DEFINE(%s_MODULE)\n", text);
@@ -1129,7 +1477,7 @@ static void output_makeconf(void)
 	char *text;
 	
 	printf("# Soundcard configuration for ALSA driver\n");
-	printf("# Copyright (c) by Jaroslav Kysela <perex@suse.cz>,\n");
+	printf("# Copyright (c) by Jaroslav Kysela <perex@perex.cz>,\n");
 	printf("#                  Anders Semb Hermansen <ahermans@vf.telia.no>\n\n");
 	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
@@ -1145,12 +1493,12 @@ static void output_include(void)
 	char *text;
 	
 	printf("/* Soundcard configuration for ALSA driver */\n");
-	printf("/* Copyright (c) by Jaroslav Kysela <perex@suse.cz>, */\n");
+	printf("/* Copyright (c) by Jaroslav Kysela <perex@perex.cz>, */\n");
 	printf("/*                  Anders Semb Hermansen <ahermans@vf.telia.no> */\n\n");
 	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
 		printf("#undef %s%s\n", text,
-		       tempdep->is_bool ? "" : "_MODULE");
+		       tempdep->type == TYPE_TRISTATE ? "_MODULE" : "");
 		free(text);
 	}
 }
@@ -1278,6 +1626,12 @@ int main(int argc, char *argv[])
 			argidx += 2;
 			continue;
 		}
+		if (strcmp(argv[argidx], "--versiondep") == 0) {
+			if (read_version_deps(argv[argidx + 1]) < 0)
+				exit(EXIT_FAILURE);
+			argidx += 2;
+			continue;
+		}
 		break;
 	}
 	if (strcmp(argv[argidx], "--acinclude") == 0)
